import os

import pkg_resources
from flask import Flask, send_from_directory

from cirrocumulus.api import blueprint, auth_api, database_api, dataset_api
from cirrocumulus.envir import CIRRO_AUTH_CLIENT_ID
from cirrocumulus.firestore_datastore import FirestoreDatastore
from cirrocumulus.google_auth import GoogleAuth
from cirrocumulus.parquet_dataset import ParquetDataset

pkg_resources.resource_filename("cirrocumulus", 'seurat2loom.R')
# If `entrypoint` is not defined in app.yaml, App Engine will look for an app
# called `app` in `main.py`.


app = Flask(__name__, static_folder='cirrocumulus/client', static_url_path='')
app.register_blueprint(blueprint, url_prefix='/api')


@app.route('/')
def root():
    return send_from_directory(os.path.abspath(os.path.join(app.root_path, "cirrocumulus", "client")), "index.html")


dataset_api.add(ParquetDataset())

try:
    from cirrocumulus.zarr_dataset_backed import ZarrDatasetBacked

    dataset_api.add(ZarrDatasetBacked())
except ModuleNotFoundError:
    pass

auth_api.provider = GoogleAuth(os.environ[CIRRO_AUTH_CLIENT_ID])
database_api.provider = FirestoreDatastore()

if __name__ == '__main__':
    # from flask_cors import CORS
    # CORS(app)
    app.run(host='127.0.0.1', port=5000, debug=True)



import gzip
import logging
import os

import numpy as np
import pyarrow as pa
import pyarrow.parquet as pq
import scipy.sparse

from cirrocumulus.simple_data import SimpleData

logger = logging.getLogger("cirro")


def write_pq(d, output_dir, name, write_statistics=True, row_group_size=None):
    os.makedirs(output_dir, exist_ok=True)
    pq.write_table(pa.Table.from_pydict(d), output_dir + os.path.sep + name + '.parquet',
        write_statistics=write_statistics, row_group_size=row_group_size)


def save_adata_pq(adata, output_directory):
    import pandas._libs.json as ujson
    logger.info('Save adata')
    X_dir = os.path.join(output_directory, 'X')
    obs_dir = os.path.join(output_directory, 'obs')
    obsm_dir = os.path.join(output_directory, 'obsm')
    result = SimpleData.schema(adata)
    result['format'] = 'parquet'
    os.makedirs(X_dir, exist_ok=True)
    os.makedirs(obs_dir, exist_ok=True)
    os.makedirs(obsm_dir, exist_ok=True)
    with gzip.open(os.path.join(output_directory, 'index.json.gz'), 'wt') as f:
        # json.dump(result, f)
        f.write(ujson.dumps(result, double_precision=2, orient='values'))

    save_adata_X(adata, X_dir)
    save_data_obs(adata, obs_dir)
    save_data_obsm(adata, obsm_dir)


def save_adata_X(adata, X_dir):
    adata_X = adata.X
    names = adata.var.index
    is_sparse = scipy.sparse.issparse(adata_X)
    if is_sparse and scipy.sparse.isspmatrix_csr(adata_X):
        adata_X = adata_X.tocsc()

    for j in range(adata_X.shape[1]):
        X = adata_X[:, j]
        if is_sparse:
            X = X.toarray().flatten()
        indices = np.where(X != 0)[0]
        values = X[indices]
        write_pq(dict(index=indices, value=values), X_dir, names[j])
        if j > 0 and (j + 1) % 1000 == 0:
            logger.info('Wrote adata X {}/{}'.format(j + 1, adata_X.shape[1]))


def save_data_obsm(adata, obsm_dir):
    logger.info('writing adata obsm')

    for name in adata.obsm.keys():
        m = adata.obsm[name]
        dimensions = 3 if m.shape[1] > 2 else 2
        d = {}
        for i in range(dimensions):
            d[name + '_' + str(i + 1)] = m[:, i].astype('float32')
        write_pq(d, obsm_dir, name)


def save_data_obs(adata, obs_dir):
    logger.info('writing adata obs')
    for name in adata.obs:
        # TODO sort?
        value = adata.obs[name]
        write_pq(dict(value=value), obs_dir, name)
    write_pq(dict(value=adata.obs.index.values), obs_dir, 'index')



import datetime
import json

from google.cloud import datastore

from cirrocumulus.database_api import get_email_domain
from .invalid_usage import InvalidUsage

DATASET = 'Dataset'
CAT_NAME = 'Cat_Name'
DATASET_FILTER = 'Dataset_Filter'
USER = 'User'


class FirestoreDatastore:

    def __init__(self):
        self.datastore_client = datastore.Client()

    def server(self):
        client = self.datastore_client
        return dict(mode='server', email=client.project + '@appspot.gserviceaccount.com')

    def category_names(self, email, dataset_id):
        dataset_id = int(dataset_id)
        self.__get_key_and_dataset(email, dataset_id, False)
        client = self.datastore_client
        query = client.query(kind=CAT_NAME)
        query.add_filter('dataset_id', '=', dataset_id)
        results = []
        for result in query.fetch():
            results.append(result)
        return results

    def upsert_category_name(self, email, category, dataset_id, original_name, new_name):
        dataset_id = int(dataset_id)
        self.__get_key_and_dataset(email, dataset_id, False)
        client = self.datastore_client
        key = client.key(CAT_NAME, str(dataset_id) + '-' + str(category) + '-' + str(original_name))

        if new_name == '':
            client.delete(key)
        else:
            entity = datastore.Entity(key=key)
            entity.update(dict(category=category, dataset_id=dataset_id, original=original_name, new=new_name))
            client.put(entity)
            return entity.id

    def user(self, email):
        client = self.datastore_client
        key = client.key(USER, email)
        user = client.get(key)
        if user is None:
            user = datastore.Entity(client.key(USER, email))
        user.update({'last_login': datetime.datetime.now()})
        client.put(user)
        return user

    def datasets(self, email):
        client = self.datastore_client
        query = client.query(kind=DATASET)
        query.add_filter('readers', '=', email)
        results = []
        for result in query.fetch():
            results.append(
                {'id': result.id, 'name': result['name'], 'owner': email in result['owners'], 'url': result['url']})
        domain = get_email_domain(email)
        if domain is not None:
            query = client.query(kind=DATASET)
            query.add_filter('readers', '=', domain)
            unique_ids = set()
            for result in results:
                unique_ids.add(result['id'])
            for result in query.fetch():
                if result.id not in unique_ids:
                    results.append({'id': result.id, 'name': result['name'], 'owner': email in result['owners'],
                                    'url': result['url']})
        return results

    def dataset_filters(self, email, dataset_id):
        client = self.datastore_client
        self.__get_key_and_dataset(email, dataset_id)
        query = client.query(kind=DATASET_FILTER)
        query.add_filter('dataset_id', '=', int(dataset_id))
        results = []
        for result in query.fetch():
            results.append(result)
        return results

    def delete_dataset_filter(self, email, dataset_id, filter_id):
        client = self.datastore_client
        key = client.key(DATASET_FILTER, int(filter_id))
        dataset_filter = client.get(key)
        self.__get_key_and_dataset(email, dataset_filter['dataset_id'])
        client.delete(key)

    def get_dataset_filter(self, email, dataset_id, filter_id):
        client = self.datastore_client
        dataset_filter = client.get(client.key(DATASET_FILTER, int(filter_id)))
        self.__get_key_and_dataset(email, dataset_filter['dataset_id'])
        return dataset_filter

    def upsert_dataset_filter(self, email, dataset_id, filter_id, filter_name, filter_notes, dataset_filter):
        dataset_id = int(dataset_id)
        client = self.datastore_client
        self.__get_key_and_dataset(email, dataset_id)
        if filter_id is None:
            dataset_filter_entity = datastore.Entity(client.key(DATASET_FILTER),
                exclude_from_indexes=['value'])
        else:
            filter_id = int(filter_id)
            key = client.key(DATASET_FILTER, filter_id)
            dataset_filter_entity = client.get(key)
        entity_update = {}
        if filter_name is not None:
            entity_update['name'] = filter_name
        if dataset_filter is not None:
            entity_update['value'] = json.dumps(dataset_filter)
        if email is not None:
            entity_update['email'] = email
        if dataset_id is not None:
            entity_update['dataset_id'] = dataset_id
        if filter_notes is not None:
            entity_update['notes'] = filter_notes

        dataset_filter_entity.update(entity_update)
        client.put(dataset_filter_entity)
        return dataset_filter_entity.id

    def __get_key_and_dataset(self, email, dataset_id, ensure_owner=False):
        client = self.datastore_client
        key = client.key(DATASET, int(dataset_id))
        dataset = client.get(key)
        if dataset is None:
            raise InvalidUsage('Please provide a valid id', 400)
        readers = dataset.get('readers')

        domain = get_email_domain(email)
        if email not in readers and domain not in readers:
            raise InvalidUsage('Not authorized', 403)
        if ensure_owner and email not in dataset['owners']:
            raise InvalidUsage('Not authorized', 403)
        return key, dataset

    def delete_dataset(self, email, dataset_id):
        client = self.datastore_client
        key, dataset = self.__get_key_and_dataset(email, dataset_id, True)
        client.delete(key)

    def get_dataset(self, email, dataset_id, ensure_owner=False):
        _, dataset = self.__get_key_and_dataset(email, dataset_id, ensure_owner)
        dataset['id'] = dataset.id
        return dataset

    def upsert_dataset(self, email, dataset_id, dataset_name, url, readers):
        client = self.datastore_client
        if dataset_id is not None:  # only owner can update
            key, dataset = self.__get_key_and_dataset(email, dataset_id, True)
        else:
            dataset = datastore.Entity(client.key(DATASET), exclude_from_indexes=['url', 'precomputed'])
            user = client.get(client.key(USER, email))
            if 'importer' not in user or not user['importer']:
                raise InvalidUsage('Not authorized', 403)
        readers = set(readers)
        if email in readers:
            readers.remove(email)
        readers.add(email)
        update_dict = {'name': dataset_name,
                       'readers': list(readers),
                       'url': url}

        if dataset_id is None:  # new dataset
            update_dict['owners'] = [email]

        dataset.update(update_dict)
        client.put(dataset)
        dataset_id = dataset.id
        return dataset_id



# import pandas.io.json as json
import pandas._libs.json as ujson
from flask import make_response


def to_json(data, response=200):
    # response = make_response(simplejson.dumps(data, check_circular=True), response)
    # response = make_response(json.dumps(data), response)
    s = ujson.dumps(data, double_precision=2, orient='values')
    # s = nujson.dumps(data, double_precision=1)
    response = make_response(s, response)
    response.headers['Content-Type'] = 'application/json'
    return response



class AuthAPI:

    def __init__(self):
        self.provider = None

    @property
    def client_id(self):
        return self.provider.client_id

    def auth(self):
        return self.provider.auth()



def get_email_domain(email):
    at_index = email.find('@')
    domain = None
    if at_index != -1:
        domain = email[at_index + 1:]
    return domain


def load_dataset_schema(url):
    import os
    from urllib.parse import urlparse
    import fsspec
    import json

    def get_extension(path):
        name, ext = os.path.splitext(path)
        if ext == '.gz':
            name, ext = os.path.splitext(name)
            if ext == '.json':
                ext = '.json.gz'
        return ext

    extension = get_extension(url)
    json_schema = None
    if extension in ['.json', '.json.gz', '']:
        pr = urlparse(url)
        fs = fsspec.filesystem(pr.scheme if not pr.scheme == '' else 'file')
        if extension == '':
            url = os.path.join(url, 'index.json.gz')
            extension = get_extension(url)
        if extension == '.json.gz':
            import gzip
            with gzip.open(fs.open(url)) as f:
                json_schema = json.load(f)
        else:
            with fs.open(url) as f:
                json_schema = json.load(f)
    return json_schema


class DatabaseAPI:

    def __init__(self):
        self.provider = None

    def server(self):
        return self.provider.server()

    def user(self, email):
        return self.provider.user(email)

    def datasets(self, email):
        return self.provider.datasets(email)

    def category_names(self, email, dataset_id):
        return self.provider.category_names(email, dataset_id)

    def upsert_category_name(self, email, category, dataset_id, original_name, new_name):
        return self.provider.upsert_category_name(email, category, dataset_id, original_name, new_name)

    def delete_dataset(self, email, dataset_id):
        return self.provider.delete_dataset(email, dataset_id)

    def get_dataset(self, email, dataset_id, ensure_owner=False):
        return self.provider.get_dataset(email, dataset_id, ensure_owner)

    def upsert_dataset(self, email, dataset_id, dataset_name, url, readers):
        return self.provider.upsert_dataset(email, dataset_id, dataset_name, url, readers)

    def dataset_filters(self, email, dataset_id):
        return self.provider.dataset_filters(email, dataset_id)

    def get_dataset_filter(self, email, dataset_id, filter_id):
        return self.provider.get_dataset_filter(email, dataset_id, filter_id)

    def upsert_dataset_filter(self, email, dataset_id, filter_id, filter_name, filter_notes, dataset_filter):
        return self.provider.upsert_dataset_filter(email, dataset_id, filter_id, filter_name, filter_notes,
            dataset_filter)

    def delete_dataset_filter(self, email, dataset_id, filter_id):
        return self.provider.delete_dataset_filter(email, dataset_id, filter_id)



import numpy as np
import pandas as pd
import scipy.sparse


class SimpleData:

    def __init__(self, X, obs, var):
        self.obs = obs
        self.var = var
        self.uns = {}
        if X is not None:
            if len(X.shape) == 1:
                X = np.array([X]).T
            n_var = X.shape[1]
            n_obs = X.shape[0]
        else:
            n_var = len(var) if var is not None else 0
            n_obs = len(obs) if obs is not None else 0
        self.X = X
        self.shape = (n_obs, n_var)

    @staticmethod
    def view(adata, row_slice):
        X = adata.X[row_slice] if adata.X is not None else None
        obs = adata.obs[row_slice] if adata.obs is not None else None
        return SimpleData(X, obs, adata.var)

    @staticmethod
    def obs_stats(df, columns):
        df = df[columns]
        # variables on columns, stats on rows, transpose so that stats are on columns
        return df.agg(['min', 'max', 'sum', 'mean']).T

    @staticmethod
    def X_stats(df, var_ids):
        df = df[var_ids]
        if len(df) == 0:
            zeros = np.full(len(var_ids), 0)
            empty = np.full(len(var_ids), np.nan)
            return pd.DataFrame(
                data={'min': empty, 'max': empty, 'sum': zeros,
                      'numExpressed': zeros,
                      'mean': empty}, index=var_ids)

        return pd.DataFrame(
            data={'min': np.min(df.values, axis=0), 'max': np.max(df.values, axis=0), 'sum': df.sum().values,
                  'numExpressed': (df.values != 0).sum(axis=0),
                  'mean': df.mean().values}, index=var_ids)

    @staticmethod
    def get_var_indices(adata, names):
        return adata.var.index.get_indexer_for(names)

    @staticmethod
    def find_markers(adata, key, marker_dict, n_genes):
        import scipy.stats as ss
        markers = {}
        marker_dict[key + ' markers'] = markers
        for cat in adata.obs[key].cat.categories:

            mask = adata.obs[key] == cat
            ds1 = adata[mask]
            ds_rest = adata[~mask]
            gene_names_keep = ds1.X.mean(axis=0) > ds_rest.X.mean(axis=0)
            ds1 = ds1[:, gene_names_keep.A1]
            ds_rest = ds_rest[:, gene_names_keep.A1]
            stats = np.zeros(ds1.shape[1], dtype=np.float32)
            pvals = np.full(ds1.shape[1], 1.0)
            for i in range(ds1.shape[1]):
                v1 = ds1.X[:, i]
                v2 = ds_rest.X[:, i]
                if v1.data.size > 0 and v2.data.size > 0:
                    stats[i], pvals[i] = ss.mannwhitneyu(v1.toarray()[:, 0], v2.toarray()[:, 0],
                        alternative="two-sided")

            order = np.argsort(pvals)
            markers[str(cat)] = ds1.var_names[order][:n_genes]

    @staticmethod
    def has_markers(adata):
        return hasattr(adata, 'uns') and 'rank_genes_groups' in adata.uns or 'de_res' in adata.varm

    @staticmethod
    def schema(adata):
        obs_cat = []
        obs = []
        result = {'version': '1.0.0'}
        marker_dict = adata.uns.get('markers', {})
        result['markers'] = marker_dict
        n_genes = 10
        if SimpleData.has_markers(adata):
            if hasattr(adata, 'uns') and 'rank_genes_groups' in adata.uns:  # scanpy
                for key in adata.uns.keys():
                    rank_genes_groups = adata.uns[key]
                    if isinstance(rank_genes_groups, dict) and 'logfoldchanges' in rank_genes_groups:
                        groupby = str(rank_genes_groups['params']['groupby'])
                        group_names = rank_genes_groups['names'].dtype.names
                        markers = {}
                        markers_key = groupby + ' markers'
                        duplicate_counter = 1
                        while markers_key in markers:
                            markers_key = groupby + ' markers-{}'.format(duplicate_counter)
                            duplicate_counter += 1
                        marker_dict[markers_key] = markers
                        for group_name in group_names:
                            gene_names = rank_genes_groups['names'][group_name]
                            # scores = rank_genes_groups['scores'][group_name]
                            markers[group_name] = gene_names[:n_genes]
            else:  # pegasus
                de_res = adata.varm['de_res']
                names = de_res.dtype.names
                base_names = [name[:name.rindex(':')] for name in names]
                fields = ['auroc', 'log_fold_change']
                field_use = None
                for field in fields:
                    if field in base_names:
                        field_use = field
                        break
                if field_use is None:
                    print('Pegasus differential expression results not found')
                else:
                    markers = {}
                    marker_dict['markers'] = markers
                    for name in names:
                        index = name.rindex(':')
                        base_name = name[:index]

                        if base_name == field_use:
                            cluster_name = name[index + 1:]
                            indices = np.argsort(de_res[name])
                            markers[cluster_name] = adata.var.index[indices[len(indices) - n_genes:]]

        for key in adata.obs_keys():
            if pd.api.types.is_categorical_dtype(adata.obs[key]) or pd.api.types.is_bool_dtype(
                    adata.obs[key]) or pd.api.types.is_object_dtype(adata.obs[key]):
                obs_cat.append(key)
            else:
                obs.append(key)
        # spatial_node = adata.uns['spatial'] if 'spatial' in adata.uns else None
        #
        # if spatial_node is not None:
        #     spatial_node_keys = list(spatial_node.keys())  # list of datasets
        #     if len(spatial_node_keys) == 1:
        #         spatial_node = spatial_node[spatial_node_keys[0]]

        images_node = adata.uns.get('images', [])
        image_names = list(map(lambda x: x['name'], images_node))
        result['var'] = adata.var_names.values
        result['obs'] = obs
        result['obsCat'] = obs_cat
        result['shape'] = adata.shape

        embeddings = []
        for key in adata.obsm_keys():
            dim = min(3, adata.obsm[key].shape[1])
            if dim < 2:
                continue
            embedding = dict(name=key, dimensions=dim)
            try:
                image_index = image_names.index(key)
                embedding['spatial'] = images_node[image_index]
            except ValueError:
                pass

            if dim == 3:
                embeddings.append(embedding)
                embedding = embedding.copy()
                embedding['dimensions'] = 2
                embeddings.append(embedding)
            else:
                embeddings.append(embedding)
        result['embeddings'] = embeddings
        return result

    @staticmethod
    def to_df(adata, obs_measures, var_measures, dimensions, basis=None):
        df = pd.DataFrame()
        obs_keys = obs_measures + dimensions
        if basis is not None:
            obs_keys += basis['coordinate_columns']

        for key in obs_keys:
            df[key] = adata.obs[key]
        indices = SimpleData.get_var_indices(adata, var_measures)
        for i in range(len(var_measures)):
            X = adata.X[:, indices[i]]
            if scipy.sparse.issparse(X):
                X = X.toarray()
            df[var_measures[i]] = X
        return df



#!/usr/bin/env python
# -*- coding: utf-8 -*-
import argparse

import sys
from cirrocumulus import launch, prepare_data, serve


def main():
    command_list = [launch, prepare_data, serve]
    parser = argparse.ArgumentParser(description='Run a cirro command')
    command_list_strings = list(map(lambda x: x.__name__[len('cirrocumulus.'):], command_list))
    parser.add_argument('command', help='The command', choices=command_list_strings)
    parser.add_argument('command_args', help='The command arguments', nargs=argparse.REMAINDER)
    my_args = parser.parse_args()
    command_name = my_args.command
    command_args = my_args.command_args
    cmd = command_list[command_list_strings.index(command_name)]
    sys.argv[0] = cmd.__file__
    cmd.main(command_args)


if __name__ == '__main__':
    main()



from urllib.parse import urlparse

import fsspec


class FileSystemAdapter:

    def __init__(self):
        # scheme can be gs, file, etc.
        self.scheme_to_fs = {}

    def get_fs(self, path):
        pr = urlparse(path)
        fs = self.scheme_to_fs.get(pr.scheme, None)
        if fs is not None:
            return fs
        fs = fsspec.filesystem(pr.scheme if not pr.scheme == '' else 'file')
        self.scheme_to_fs[pr.scheme] = fs
        return fs



import json

from bson import ObjectId
from pymongo import MongoClient

from cirrocumulus.database_api import get_email_domain
from .invalid_usage import InvalidUsage


class MongoDb:

    def __init__(self, db_uri='mongodb://localhost:27017/', database='cirrocumulus', email=None):
        self.client = MongoClient(db_uri)
        self.db = self.client[database]
        self.email = email

    def server(self):
        d = dict(mode='server')
        if self.email is not None:
            d['email'] = self.email
        return d

    def category_names(self, email, dataset_id):
        self.get_dataset(email, dataset_id)
        collection = self.db.categories
        results = []
        for doc in collection.find(dict(dataset_id=dataset_id)):
            results.append({'category': doc['category'], 'dataset_id': doc['dataset_id'], 'original': doc['original'],
                            'new': doc['new']})
        return results

    def upsert_category_name(self, email, category, dataset_id, original_name, new_name):
        self.get_dataset(email, dataset_id)
        collection = self.db.categories
        key = str(dataset_id) + '-' + str(category) + '-' + str(original_name)

        if new_name == '':
            collection.delete_one(dict(cat_id=key))
        else:
            collection.update_one(dict(cat_id=key),
                {'$set': dict(category=category, dataset_id=dataset_id, original=original_name, new=new_name)},
                upsert=True)
            return key

    def user(self, email):
        collection = self.db.users
        doc = collection.find_one(dict(email=email))
        if doc is None:
            collection.insert_one(dict(email=email))
            return {'id': email}
        else:
            return {'id': email,
                    'importer': doc.get('importer', False)}

    def get_dataset(self, email, dataset_id, ensure_owner=False):
        collection = self.db.datasets
        doc = collection.find_one(dict(_id=ObjectId(dataset_id)))
        if doc is None:
            raise InvalidUsage('Please provide a valid id', 400)
        readers = doc.get('readers')
        domain = get_email_domain(email)
        if email not in readers and domain not in readers:
            raise InvalidUsage('Not authorized', 403)
        if ensure_owner and email not in doc['owners']:
            raise InvalidUsage('Not authorized', 403)
        return {
                'id': str(doc['_id']),
                'name': doc['name'],
                'readers': doc['readers'],
                'url': doc['url'],
                'owner': 'owners' in doc and email in doc['owners']}

    def datasets(self, email):
        collection = self.db.datasets
        results = []
        domain = get_email_domain(email)
        if domain is None:
            query = dict(readers=email)
        else:
            query = dict(readers={'$in': [email, domain]})
        for doc in collection.find(query):
            results.append({'id': str(doc['_id']), 'name': doc['name'],
                            'owner': 'owners' in doc and email in doc['owners'], 'url': doc['url']})
        return results

    def dataset_filters(self, email, dataset_id):
        self.get_dataset(email, dataset_id)
        collection = self.db.filters
        results = []

        for doc in collection.find(dict(dataset_id=dataset_id)):
            results.append(
                {'id': str(doc['_id']), 'dataset_id': doc['dataset_id'], 'name': doc['name'], 'value': doc['value'],
                 'notes': doc['notes'], 'email': doc['email']})
        return results

    def delete_dataset_filter(self, email, dataset_id, filter_id):
        collection = self.db.filters
        doc = collection.find_one(dict(_id=ObjectId(filter_id)))
        self.get_dataset(email, doc['dataset_id'])
        collection.delete_one(dict(_id=ObjectId(filter_id)))

    def get_dataset_filter(self, email, dataset_id, filter_id):
        collection = self.db.filters
        doc = collection.find_one(dict(_id=ObjectId(filter_id)))
        self.get_dataset(email, doc['dataset_id'])
        return {'id': str(doc['_id']), 'dataset_id': doc['dataset_id'], 'name': doc['name'], 'value': doc['value'],
                'notes': doc['notes'], 'email': doc['email']}

    def upsert_dataset_filter(self, email, dataset_id, filter_id, filter_name, filter_notes, dataset_filter):
        self.get_dataset(email, dataset_id)
        collection = self.db.filters
        entity_update = {}
        if filter_name is not None:
            entity_update['name'] = filter_name
        if dataset_filter is not None:
            entity_update['value'] = json.dumps(dataset_filter)
        if email is not None:
            entity_update['email'] = email
        if dataset_id is not None:
            entity_update['dataset_id'] = dataset_id
        if filter_notes is not None:
            entity_update['notes'] = filter_notes
        if filter_id is None:
            return str(collection.insert_one(entity_update).inserted_id)
        else:
            collection.update_one(dict(_id=ObjectId(filter_id)), {'$set': entity_update})
            return filter_id

    def delete_dataset(self, email, dataset_id):
        self.get_dataset(email, dataset_id, True)
        collection = self.db.datasets
        collection.delete_one(dict(_id=ObjectId(dataset_id)))
        self.db.filters.delete_many(dict(dataset_id=dataset_id))
        self.db.categories.delete_many(dict(dataset_id=dataset_id))

    def upsert_dataset(self, email, dataset_id, dataset_name, url, readers):
        collection = self.db.datasets
        readers = set(readers)
        if email in readers:
            readers.remove(email)
        readers.add(email)
        update_dict = {'name': dataset_name,
                       'readers': list(readers),
                       'url': url}
        if dataset_id is None:  # new dataset
            if email != '':
                user = self.db.users.find_one(dict(email=email))
                if 'importer' not in user or not user['importer']:
                    raise InvalidUsage('Not authorized', 403)
            update_dict['owners'] = [email]
            return str(collection.insert_one(update_dict).inserted_id)
        else:
            self.get_dataset(email, dataset_id, True)
            collection.update_one(dict(_id=ObjectId(dataset_id)), {'$set': update_dict})
            return dataset_id



import pandas as pd


class UniqueAggregator:

    def __init__(self, column):
        self.column = column

    def execute(self, df):
        # column can be cell index or bin number
        if self.column == 'index':
            values = df.index
        else:
            values = df[self.column].unique()
            if isinstance(values, pd.arrays.SparseArray):
                values = values.to_dense()
            values.sort()
        return {'indices_or_bins': values}



import pandas as pd
from natsort import natsorted
from pandas import CategoricalDtype


class DotPlotAggregator:

    def __init__(self, var_measures, dimensions):
        self.var_measures = var_measures
        self.dimensions = dimensions

    def execute(self, df):
        results = []
        # {categories:[], name:'', values:[{name:'', fractionExpressed:0, mean:0}]}
        var_measures = self.var_measures
        dimensions = self.dimensions
        if len(var_measures) == 0 or len(dimensions) == 0:
            return results

        def mean(x):
            return x.mean()

        def fraction_expressed(x):
            return (x.values != 0).sum() / len(x)

        for dimension in dimensions:
            if pd.api.types.is_categorical_dtype(df[dimension]):
                if not df[dimension].dtype.ordered:
                    df[dimension] = df[dimension].astype(
                        CategoricalDtype(natsorted(df[dimension].dtype.categories), ordered=True))
                if len(df[dimension].dtype.categories) <= 1:
                    continue
            agg_result = df.groupby(dimension, observed=True).agg([mean, fraction_expressed])

            values = []
            dotplot_result = {'categories': agg_result.index, 'name': dimension, 'values': values}
            for name in var_measures:
                series = agg_result[name]
                is_sparse = hasattr(series, 'sparse')
                if is_sparse:
                    series = series.sparse.to_dense()
                values.append({'name': name,
                               'fractionExpressed': series['fraction_expressed'],
                               'mean': series['mean']})
            results.append(dotplot_result)
        return results



from flask import Blueprint, Response, request, stream_with_context

import cirrocumulus.data_processing as data_processing
from .auth_api import AuthAPI
from .database_api import DatabaseAPI
from .dataset_api import DatasetAPI
from .invalid_usage import InvalidUsage
from .util import *


blueprint = Blueprint('blueprint', __name__)

dataset_api = DatasetAPI()

auth_api = AuthAPI()
database_api = DatabaseAPI()


@blueprint.errorhandler(InvalidUsage)
def handle_invalid_usage(error):
    return Response(error.message, error.status_code)


@blueprint.route('/server', methods=['GET'])
def handle_server():
    # no login required
    server = database_api.server()
    server['clientId'] = auth_api.client_id
    return to_json(server)


@blueprint.route('/filters', methods=['GET'])
def handle_dataset_filters():
    """List filters available for a dataset.
    """
    email = auth_api.auth()['email']
    dataset_id = request.args.get('id', '')

    if dataset_id == '':
        return 'Please provide an id', 400

    dataset_filters = database_api.dataset_filters(email, dataset_id)
    # {'id': result.id, 'name': result['name']}
    return to_json(dataset_filters)


@blueprint.route('/export_filters', methods=['GET'])
def handle_export_dataset_filters():
    """Download filters in a csv file for a dataset.
    """
    email = auth_api.auth()['email']
    dataset_id = request.args.get('id', '')

    if dataset_id == '':
        return 'Please provide an id', 400

    dataset_filters = database_api.dataset_filters(email, dataset_id)
    dataset = database_api.get_dataset(email, dataset_id)
    text = data_processing.handle_export_dataset_filters(dataset_api, dataset, dataset_filters)
    return Response(text, mimetype='text/plain')


@blueprint.route('/category_name', methods=['GET', 'PUT'])
def handle_category_name():
    """CRUD for category name.
    """

    email = auth_api.auth()['email']
    if request.method == 'GET':
        dataset_id = request.args.get('id', '')
        if dataset_id == '':
            return 'Please provide an id', 400
        return to_json(database_api.category_names(email, dataset_id))
    content = request.get_json(force=True, cache=False)
    category = content.get('c')
    original_name = content.get('o')
    new_name = content.get('n')
    dataset_id = content.get('id')
    if request.method == 'PUT':
        database_api.upsert_category_name(
            email=email,
            category=category,
            dataset_id=dataset_id,
            original_name=original_name,
            new_name=new_name)
        return '', 200


@blueprint.route('/filter', methods=['GET', 'POST', 'PUT', 'DELETE'])
def handle_dataset_filter():
    """CRUD for a dataset filter.
    """

    email = auth_api.auth()['email']
    if request.method == 'GET':
        filter_id = request.args.get('id', '')
        dataset_id = request.args.get('ds_id', '')
        if filter_id == '' or dataset_id == '':
            return 'Please provide an id', 400
        return to_json(database_api.get_dataset_filter(email, dataset_id=dataset_id, filter_id=filter_id))
    content = request.get_json(force=True, cache=False)
    filter_id = content.get('id')
    dataset_id = content.get('ds_id')
    # POST=new, PUT=update , DELETE=delete, GET=get
    if request.method == 'PUT' or request.method == 'POST':
        if request.method == 'PUT' and filter_id is None:
            return 'Please supply an id', 400
        if request.method == 'POST' and dataset_id is None:
            return 'Please supply a ds_id', 400
        dataset_filter = content.get('value')
        filter_name = content.get('name')
        filter_notes = content.get('notes')
        filter_id = database_api.upsert_dataset_filter(
            email=email,
            dataset_id=dataset_id,
            filter_id=filter_id if request.method == 'PUT' else None,
            dataset_filter=dataset_filter,
            filter_name=filter_name,
            filter_notes=filter_notes)
        return to_json({'id': filter_id})
    elif request.method == 'DELETE':
        database_api.delete_dataset_filter(email, dataset_id=dataset_id, filter_id=filter_id)
        return to_json('', 204)


@blueprint.route('/schema', methods=['GET'])
def handle_schema():
    """Get dataset schema.
    """
    email = auth_api.auth()['email']
    dataset_id = request.args.get('id', '')
    dataset = database_api.get_dataset(email, dataset_id)
    schema = dataset_api.schema(dataset)
    return to_json(schema)


@blueprint.route('/file', methods=['GET'])
def handle_file():
    email = auth_api.auth()['email']
    dataset_id = request.args.get('id', '')
    dataset = database_api.get_dataset(email, dataset_id)
    file = request.args.get('file')
    url = dataset['url']
    import os
    import mimetypes
    file_path = os.path.join(os.path.dirname(url), os.path.abspath(file))
    mimetype = mimetypes.guess_type(file_path)
    # with dataset_api.fs_adapter.get_fs(file_path).open(file_path) as f:
    #     bytes = f.read()
    # return Response(bytes, mimetype=mimetype[0])
    chunk_size = 4096
    f = dataset_api.fs_adapter.get_fs(file_path).open(file_path)

    def generate():
        while True:
            chunk = f.read(chunk_size)
            if len(chunk) <= 0:
                f.close()
                break
            yield chunk

    response = Response(stream_with_context(generate()), mimetype=mimetype[0])
    return response


@blueprint.route('/user', methods=['GET'])
def handle_user():
    email = auth_api.auth()['email']
    user = database_api.user(email)
    return to_json(user)


def get_email_and_dataset(content):
    email = auth_api.auth()['email']
    dataset_id = content.get('id', '')
    if dataset_id == '':
        return 'Please supply an id', 400
    dataset = database_api.get_dataset(email, dataset_id)
    return email, dataset


@blueprint.route('/data', methods=['POST'])
def handle_data():
    json_request = request.get_json(cache=False)
    email, dataset = get_email_and_dataset(json_request)

    return to_json(
        data_processing.handle_data(dataset_api=dataset_api, dataset=dataset,
            embedding_list=json_request.get('embedding'),
            stats=json_request.get('stats'),
            grouped_stats=json_request.get('groupedStats'),
            selection=json_request.get('selection')))


@blueprint.route('/selection', methods=['POST'])
def handle_selection():
    # selection includes stats, coordinates
    content = request.get_json(cache=False)
    email, dataset = get_email_and_dataset(content)
    data_filter = content.get('filter')
    return to_json(data_processing.handle_selection(dataset_api=dataset_api, dataset=dataset, data_filter=data_filter,
        measures=content.get('measures', []), dimensions=content.get('dimensions', []),
        embeddings=content.get('embeddings', [])))


@blueprint.route('/selected_ids', methods=['POST'])
def handle_selected_ids():
    content = request.get_json(cache=False)
    email, dataset = get_email_and_dataset(content)
    data_filter = content.get('filter')
    return to_json(
        data_processing.handle_selection_ids(dataset_api=dataset_api, dataset=dataset, data_filter=data_filter))


# List available datasets
@blueprint.route('/datasets', methods=['GET'])
def handle_datasets():
    email = auth_api.auth()['email']
    return to_json(database_api.datasets(email))


@blueprint.route('/dataset', methods=['GET', 'POST', 'PUT', 'DELETE'])
def handle_dataset():
    email = auth_api.auth()['email']
    # POST=new dataset, PUT=update dataset, DELETE=delete, GET=get dataset info
    if request.method == 'PUT' or request.method == 'POST':
        content = request.get_json(force=True, cache=False)
        dataset_id = content.get('id', '')
        if request.method == 'POST' and dataset_id == '':
            return 'Please supply an id', 400
        readers = set(content.get('readers', []))
        dataset_name = content.get('name', '')
        url = content.get('url', '')  # e.g. gs://foo/a/b/
        if dataset_name == '' or url == '':
            return 'Must supply dataset name and URL', 400
        dataset_id = database_api.upsert_dataset(email=email,
            dataset_id=dataset_id if request.method == 'PUT' else None,
            dataset_name=dataset_name, url=url, readers=readers)
        return to_json({'id': dataset_id})
    elif request.method == 'DELETE':
        content = request.get_json(force=True, cache=False)
        dataset_id = content.get('id', '')
        database_api.delete_dataset(email, dataset_id)
        return to_json('', 204)
    elif request.method == 'GET':
        dataset_id = request.args.get('id', '')
        return to_json(database_api.get_dataset(email, dataset_id, True))


    # if request.method == 'POST' or (request.method == 'PUT' and url != dataset_dict.get('url',
    #         '')):  # only check if can read on new dataset created
    # bucket = url[5:]  # remove gs://
    # slash = bucket.index('/')
    # gcp_object = urllib.parse.quote(bucket[slash + 1:], safe='')
    # bucket = urllib.parse.quote(bucket[0: slash], safe='')

    # head_request = requests.head(
    #     'https://www.googleapis.com/storage/v1/b/{bucket}/o/{object}'.format(bucket=bucket,
    #         object=gcp_object),
    #     headers={'Authorization': 'Bearer ' + request_util.credentials.get_access_token().access_token})
    # head_request.close()
    # if head_request.status_code != 200:
    #     return 'Not authorized to read {}'.format(url), 403



import argparse
import gzip
import json
import logging
import os

import anndata
import cirrocumulus.data_processing as data_processing
import numpy as np
import pandas as pd
import pandas._libs.json as ujson
from cirrocumulus.anndata_dataset import AnndataDataset
from cirrocumulus.dataset_api import DatasetAPI
from cirrocumulus.io_util import get_markers, filter_markers, add_spatial, SPATIAL_HELP
from cirrocumulus.simple_data import SimpleData
from natsort import natsorted
from pandas import CategoricalDtype

logger = logging.getLogger("cirro")


def to_json(data):
    return ujson.dumps(data, double_precision=2, orient='values')


def make_unique(index, join='-1'):
    index = index.str.replace('/', '_')
    lower_index = index.str.lower()
    if lower_index.is_unique:
        return index
    from collections import defaultdict

    indices_dup = lower_index.duplicated(keep="first")
    values_dup_lc = lower_index.values[indices_dup]
    values_dup = index.values[indices_dup]
    counter = defaultdict(lambda: 0)
    for i, v in enumerate(values_dup_lc):
        counter[v] += 1
        values_dup[i] += join + str(counter[v])
    values = index.values
    values[indices_dup] = values_dup
    index = pd.Index(values)
    return index


def make_ordered(df, columns):
    if columns is None:
        columns = df.columns
    for i in range(len(columns)):
        name = columns[i]
        c = df[name]
        if pd.api.types.is_categorical_dtype(c) and not c.dtype.ordered:
            df[name] = df[name].astype(CategoricalDtype(natsorted(df[name].dtype.categories), ordered=True))


def write_obs_stats(directory, results):
    os.makedirs(os.path.join(directory, 'stats', 'obs'), exist_ok=True)
    for column in results:
        dimension_or_measure_summary = results[column]
        if 'categories' in dimension_or_measure_summary:
            with gzip.open(os.path.join(directory, 'stats', 'obs', column + '.json.gz'), 'wt') as f:
                f.write(to_json(dimension_or_measure_summary))
        else:
            with gzip.open(os.path.join(directory, 'stats', 'obs', column + '.json.gz'), 'wt') as f:
                f.write(to_json(dimension_or_measure_summary))


def write_X_bins(base_output_dir, n_bins=25):
    # read in means from stats/X
    obs_avg = []
    names = []
    d = os.path.join(base_output_dir, 'stats', 'X')
    for file in os.listdir(d):
        if file.endswith('.json.gz'):
            with gzip.open(os.path.join(d, file), 'rt') as f:
                names.append(file[0:file.rindex('.json.gz')])
                result = json.load(f)
                obs_avg.append(result['mean'])

    obs_avg = pd.Series(np.array(obs_avg), index=names)
    n_items = int(np.round(len(obs_avg) / (n_bins - 1)))
    obs_cut = obs_avg.rank(method='min') // n_items
    obs_cut = obs_cut.astype(int)
    with gzip.open(os.path.join(base_output_dir, 'X.json.gz'), 'wt') as f:
        f.write(to_json(obs_cut.to_dict()))


def write_X_stats(directory, results):
    os.makedirs(os.path.join(directory, 'stats', 'X'), exist_ok=True)
    for column in results:
        measure_summary = results[column]
        with gzip.open(os.path.join(directory, 'stats', 'X', column + '.json.gz'), 'wt') as f:
            f.write(to_json(measure_summary))


def write_grouped_stats(directory, dotplot_results):
    # mean and fraction expressed for every gene per category
    for dotplot_result in dotplot_results:
        categories = dotplot_result['categories']
        category_name = dotplot_result['name']
        os.makedirs(os.path.join(directory, 'grouped_stats', 'obs', category_name, 'X'), exist_ok=True)
        with gzip.open(os.path.join(directory, 'grouped_stats', 'obs', category_name, 'index.json.gz'),
                'wt') as f:
            f.write(to_json(categories))
        values = dotplot_result['values']
        for value in values:
            var_field = value['name']
            with gzip.open(os.path.join(directory, 'grouped_stats', 'obs', category_name, 'X', var_field + '.json.gz'),
                    'wt') as f:
                f.write(to_json(value))


def write_basis_X(coords_group, basis_group, adata, result):
    shape = (coords_group['index'].shape[0], adata.shape[1])
    X = basis_group.require_dataset('X', shape=shape, chunks=(None, 1), dtype='f4')
    for column in result['values']:
        X_index = adata.var.index.get_indexer_for([column])[0]
        X[:, X_index] = result['values'][column]


def require_binned_basis_group(store, basis):
    basis_group = store.require_group('obsm_summary/' + basis['full_name'])
    basis_group.attrs['nbins'] = basis['nbins']
    basis_group.attrs['agg'] = basis['agg']
    basis_group.attrs['name'] = basis['name']
    basis_group.attrs['dimensions'] = basis['dimensions']
    basis_group.attrs['precomputed'] = True
    return basis_group


def write_basis_obs(basis, coords_group, obs_group, result):
    coords_group['index'] = result['bins']
    ndim = len(basis['coordinate_columns'])
    nbins = len(result['bins'])

    value_ds = coords_group.require_dataset('value', dtype='f4', shape=(nbins, ndim))

    for i in range(len(basis['coordinate_columns'])):
        coord_field = basis['coordinate_columns'][i]
        value_ds[:, i] = np.array(result['coordinates'][coord_field])
    # write_pq(bin_dict, output_directory, 'index')

    for column in result['values']:
        vals = result['values'][column]
        if not isinstance(vals, dict):  # continuous field
            obs_group[column] = vals
        else:
            g = obs_group.require_group(column)
            g['purity'] = vals['purity']
            g['mode'] = vals['value']
        # write_pq(vals, output_directory, column)


class PrepareData:

    def __init__(self, input_path, backed, output, X_range=None, dimensions=None, groups=[], markers=[],
                 spatial_directory=None):
        self.input_path = input_path
        self.groups = groups
        self.markers = markers
        self.stats = False
        self.nbins = None
        self.bin_agg_function = None
        # if basis_list is None or len(basis_list) == 0:
        #     basis_list = list(self.adata.obsm_keys())
        self.basis_list_to_precompute = None
        self.adata = anndata.read_loom(input_path) if input_path.lower().endswith('.loom') else anndata.read(input_path,
            backed=backed)
        # if 'seurat_clusters' in self.adata.obs:
        #     self.adata.obs['seurat_clusters'] = self.adata.obs['seurat_clusters'].astype('category')
        if spatial_directory is not None:
            if not add_spatial(self.adata, spatial_directory):
                print('No spatial data found in {}'.format(spatial_directory))
        if not backed:
            sums = self.adata.X.sum(axis=0)
            if isinstance(sums, np.matrix):
                sums = sums.A1
            self.adata = self.adata[:, sums > 0]
        index = make_unique(self.adata.var.index.append(pd.Index(self.adata.obs.columns)))
        self.adata.var.index = index[0:len(self.adata.var.index)]
        self.adata.obs.columns = index[len(self.adata.var.index):]

        logger.info('{} - {} x {}'.format(input_path, self.adata.shape[0], self.adata.shape[1]))
        self.X_range = (0, self.adata.shape[1]) if X_range is None else X_range
        self.base_output_dir = output

        self.dataset_api = DatasetAPI()
        ds = AnndataDataset()
        ds.path_to_data[input_path] = self.adata
        self.dataset_api.add(ds)

        self.input_dataset = {'name': os.path.splitext(os.path.basename(input_path))[0], 'url': input_path,
                              'id': input_path}
        dimensions_supplied = dimensions is not None and len(dimensions) > 0
        self.dimensions = [] if not dimensions_supplied else dimensions
        self.measures = []
        self.others = []
        for i in range(len(self.adata.obs.columns)):
            name = self.adata.obs.columns[i]
            c = self.adata.obs[name]
            if pd.api.types.is_object_dtype(c):
                self.adata.obs[name] = self.adata.obs[name].astype('category')
                c = self.adata.obs[name]
            if not dimensions_supplied and pd.api.types.is_categorical_dtype(c):
                if 1 < len(c.cat.categories) < 2000:
                    self.dimensions.append(name)
                    if c.isna().sum() > 0:
                        logger.info('Replacing nans in {}'.format(name))
                        self.adata.obs[name] = self.adata.obs[name].astype(str)
                        self.adata.obs.loc[self.adata.obs[name].isna(), name] = ''
                        self.adata.obs[name] = self.adata.obs[name].astype('category')
                else:
                    self.others.append(name)
            elif not pd.api.types.is_string_dtype(c) and not pd.api.types.is_object_dtype(c):
                self.measures.append('obs/' + name)
            else:
                self.others.append(name)
        make_ordered(self.adata.obs, self.dimensions)

    def get_path(self, path):
        return os.path.join(self.base_output_dir, path)

    def execute(self):

        basis_list = self.basis_list_to_precompute
        nbins = self.nbins
        bin_agg_function = self.bin_agg_function
        X_range = self.X_range

        marker_dict = self.adata.uns.get('markers', {})
        self.adata.uns['markers'] = marker_dict
        if self.groups is not None:
            n_genes = 10
            for field in self.groups:
                if len(self.adata.obs[field].cat.categories) > 1:
                    logger.info('Computing markers for {}'.format(field))
                    SimpleData.find_markers(self.adata, field, marker_dict, n_genes)
        output_format = 'parquet'
        if X_range[0] == 0:
            if output_format == 'parquet':
                from cirrocumulus.parquet_io import save_adata_pq
                save_adata_pq(self.adata, self.base_output_dir)
            elif output_format == 'json':
                from cirrocumulus.json_io import save_adata_json
                save_adata_json(self.adata, self.base_output_dir, False)

        # if self.stats:
        #     self.summary_stats()
        #     self.grouped_stats()

        if nbins is not None and basis_list is not None:
            for basis_name in basis_list:
                self.grid_embedding(basis_name, bin_agg_function, nbins)
        self.schema()

    def summary_stats(self):
        dimensions = self.dimensions
        measures = self.measures
        dataset_api = self.dataset_api
        input_dataset = self.input_dataset
        X_range = self.X_range
        adata = self.adata
        base_dir = self.base_output_dir

        if X_range[0] == 0:
            process_results = data_processing.handle_data(dataset_api=dataset_api, dataset=input_dataset,
                stats=dict(measures=measures, dimensions=dimensions))
            write_obs_stats(base_dir, process_results['summary'])

        logger.info('Summary stats X {}-{}'.format(X_range[0], X_range[1]))
        process_results = data_processing.handle_data(dataset_api=dataset_api, dataset=input_dataset,
            stats=dict(measures=adata.var_names[X_range[0]:X_range[1]], dimensions=[]))
        write_X_stats(base_dir, process_results['summary'])
        write_X_bins(base_dir)

    # stats, grouped by categories in adata.obs for dot plot
    def grouped_stats(self):
        dimensions = self.dimensions
        dataset_api = self.dataset_api
        input_dataset = self.input_dataset
        adata = self.adata
        X_range = self.X_range
        base_dir = self.base_output_dir
        logger.info('Grouped stats X {}-{}'.format(X_range[0], X_range[1]))
        process_results = data_processing.handle_data(dataset_api=dataset_api, dataset=input_dataset,
            grouped_stats=dict(measures=adata.var_names[X_range[0]:X_range[1]], dimensions=dimensions))
        dotplot_results = process_results['dotplot']

        write_grouped_stats(base_dir, dotplot_results)


    # def grid_embedding(self, basis_name, summary, nbins):
    #     if nbins <= 0:
    #         return
    #     logger.info('{} embedding'.format(basis_name))
    #     basis = get_basis(basis_name, nbins, summary, 2, False)
    #
    #     dimensions = self.dimensions
    #     measures = self.measures
    #     dataset_api = self.dataset_api
    #     input_dataset = self.input_dataset
    #     adata = self.adata
    #     X_range = self.X_range
    #     full_basis_name = basis['full_name']
    #
    #     basis_group = require_binned_basis_group(store, basis)
    #     obs_group = basis_group.require_group('obs')
    #     coords_group = basis_group.require_group('coords')
    #
    #     if X_range[0] == 0:
    #         df_with_coords = pd.DataFrame()
    #         obsm = adata.obsm[basis_name]
    #         for i in range(len(basis['coordinate_columns'])):
    #             df_with_coords[basis['coordinate_columns'][i]] = obsm[:, i]
    #         EmbeddingAggregator.convert_coords_to_bin(df_with_coords, nbins, basis['coordinate_columns'],
    #             bin_name=full_basis_name)
    #         coords_group['bins'] = df_with_coords[full_basis_name].values
    #         coords_group['bin_coords'] = df_with_coords[basis['coordinate_columns']].values
    #
    #         result = data_processing.handle_data(dataset_api=dataset_api, dataset=input_dataset, basis=basis,
    #             measures=measures + ['__count'], dimensions=dimensions, quick=False)
    #         write_basis_obs(basis, coords_group, obs_group, result)
    #         logger.info('{} embedding finished writing obs'.format(basis_name))
    #     # write X to obsm_summary/name
    #     logger.info('{} embedding X {}-{}'.format(basis_name, X_range[0], X_range[1]))
    #     result = data_processing.handle_embedding(dataset_api=dataset_api, dataset=input_dataset, basis=basis,
    #         measures=adata.var_names[X_range[0]:X_range[1]], dimensions=[])
    #     write_basis_X(coords_group, basis_group, adata, result)
    #     # write_pq(dict(value=result['values'][column]), output_directory, column)

    def schema(self):
        basis_list = self.basis_list_to_precompute
        nbins = self.nbins
        bin_agg_function = self.bin_agg_function
        X_range = self.X_range
        if X_range[0] == 0:
            output_file = os.path.join(self.base_output_dir, 'index.json.gz')
            result = SimpleData.schema(self.adata)
            marker_dict = self.adata.uns.get('markers', {})
            if self.markers is not None:
                marker_dict.update(get_markers(self.markers))
            result['markers'] = filter_markers(self.adata, marker_dict)
            result['format'] = 'parquet'
            if self.stats:
                result['precomputed'] = True  # has stats
            if nbins is not None:
                result['embeddings'] = []
                for basis_name in basis_list:
                    result['embeddings'].append(
                        {'precomputed': True, 'name': basis_name, 'nbins': nbins, 'agg': bin_agg_function,
                         'dimensions': 2})
            with gzip.open(output_file, 'wt') as f:
                f.write(to_json(result))


def main(argsv):
    parser = argparse.ArgumentParser(
        description='Prepare a dataset for cirrocumulus server.')
    parser.add_argument('dataset', help='Path to a h5ad, loom, or Seurat file')
    parser.add_argument('--out', help='Path to output directory')
    # parser.add_argument('--stats', dest="stats", help='Generate precomputed stats', action='store_true')
    parser.add_argument('--backed', help='Load h5ad file in backed mode', action='store_true')
    # parser.add_argument('--basis', help='List of embeddings to precompute', action='append')
    #  parser.add_argument('--groups', help='List of groups to precompute summary statistics', action='append')
    parser.add_argument('--markers',
        help='Path to JSON file that maps name to features. For example {"a":["gene1", "gene2"], "b":["gene3"]',
        action='append')
    parser.add_argument('--groups', help='List of groups to compute markers for (e.g. louvain)', action='append')
    parser.add_argument('--spatial', help=SPATIAL_HELP)

    # parser.add_argument('--nbins', help='Number of bins. Set to 0 to disable binning', default=500, type=int)
    # parser.add_argument('--summary', help='Bin summary statistic for numeric values', default='max')
    # parser.add_argument('--X_range', help='Start and end position of data matrix (e.g. 0-5000)', type=str)

    args = parser.parse_args(argsv)
    logger.setLevel(logging.DEBUG)
    logger.addHandler(logging.StreamHandler())

    input_basis = []  # check if specified as comma delimited list
    input_X_range = None
    out = args.out
    input_dataset = args.dataset
    if out is None:
        out = os.path.basename(input_dataset)
        out = out[0:out.rindex('.')]
    loom_file = None
    if input_dataset.lower().endswith('.rds'):
        import subprocess
        import tempfile
        import pkg_resources

        _, loom_file = tempfile.mkstemp(suffix='.loom')
        os.remove(loom_file)
        subprocess.check_call(
            ['Rscript', pkg_resources.resource_filename("cirrocumulus", 'seurat2loom.R'), input_dataset, loom_file])
        input_dataset = loom_file
    summary = 'max'
    nbins = -1
    # summary = args.summary
    # if args.basis is not None:
    #     for b in args.basis:
    #         input_basis.extend(b.split(','))
    # input_X_range = args.X_range
    # if input_X_range is not None:
    #     input_X_range = input_X_range.split('-')
    #     assert len(input_X_range) == 2
    #     input_X_range[0] = int(input_X_range[0])
    #     input_X_range[1] = int(input_X_range[1])

    prepare_data = PrepareData(input_path=input_dataset, backed=args.backed, output=out,
        dimensions=args.groups, groups=args.groups, markers=args.markers, spatial_directory=args.spatial)
    prepare_data.execute()
    if loom_file is not None:
        os.remove(loom_file)


if __name__ == '__main__':
    import sys

    main(sys.argv)



import os

import anndata
import pandas as pd

SPATIAL_HELP = 'Directory containing 10x visium spatial data (tissue_hires_image.png, scalefactors_json.json, and tissue_positions_list.csv) ' + 'or a directory containing `image.png`, `positions.image.csv` with headers barcode, x, and y, and optionally `diameter.image.txt` containing spot diameter',


def __add_visium(adata, spatial_directory):
    scale_factors_path = os.path.join(spatial_directory, 'scalefactors_json.json')
    tissue_hires_image_path = os.path.join(spatial_directory, 'tissue_hires_image.png')
    tissue_positions_list_path = os.path.join(spatial_directory, 'tissue_positions_list.csv')
    is_visium = True
    for path in [scale_factors_path, tissue_hires_image_path, tissue_positions_list_path]:
        if not os.path.exists(path):
            is_visium = False
            break
    if is_visium:
        import json
        with open(os.path.join(spatial_directory, 'scalefactors_json.json'), 'rt') as f:
            scalefactors = json.load(f)
            # {"spot_diameter_fullres": 89.49502418224989, "tissue_hires_scalef": 0.17011142,
            # "fiducial_diameter_fullres": 144.56888521748058, "tissue_lowres_scalef": 0.051033426}
        # barcode, in_tissue, array_row, array_col, pxl_col_in_fullres, pxl_row_in_fullres
        positions = pd.read_csv(tissue_positions_list_path, header=None)
        positions.columns = [
                'barcode',
                'in_tissue',
                'array_row',
                'array_col',
                'pxl_col_in_fullres',
                'pxl_row_in_fullres',
        ]
        positions.index = positions['barcode']
        positions = positions.reindex(adata.obs.index)
        spatial_coords = positions[['pxl_row_in_fullres', 'pxl_col_in_fullres']].values
        adata.obsm['tissue_hires'] = spatial_coords * scalefactors['tissue_hires_scalef']
        adata.uns['images'] = [dict(name='tissue_hires', image=tissue_hires_image_path,
            spot_diameter=scalefactors['spot_diameter_fullres'] * scalefactors['tissue_hires_scalef'])]
        return True
    else:
        return False


def __add_generic_spatial(adata, spatial_directory):
    # positions.image.csv with barcode, x, y
    # image.png
    # diameter.image.txt (optional)
    image_extensions = set(['.png', '.jpeg', '.jpg'])
    found = False
    for f in os.listdir(spatial_directory):
        name, ext = os.path.splitext(f)
        ext = ext.lower()
        if ext in image_extensions:
            positions_path = os.path.join(spatial_directory, 'positions.' + name + '.csv')

            if os.path.exists(positions_path):
                diameter_path = os.path.join(spatial_directory, 'diameter.' + name + '.txt')
                spot_diameter = None
                if os.path.exists(diameter_path):
                    with open(diameter_path, 'rt') as diameter_in:
                        spot_diameter = float(diameter_in.readline().strip())

                positions = pd.read_csv(positions_path, index_col='barcode')
                if not pd.api.types.is_string_dtype(positions.index):
                    positions.index = positions.index.astype(str)
                positions = positions.reindex(adata.obs.index)
                adata.obsm[name] = positions[['x', 'y']].values
                adata.uns['images'] = [dict(name=name, image=os.path.join(spatial_directory, f),
                    spot_diameter=spot_diameter)]
                found = True
    return found


def add_spatial(adata, spatial_directory):
    if not __add_visium(adata, spatial_directory):
        return __add_generic_spatial(adata, spatial_directory)
    else:
        return True


def get_cell_type_genes(cell_type):
    all_genes = []

    for cell_type_markers in cell_type['markers']:
        all_genes += cell_type_markers['genes']

    for i in range(len(all_genes)):
        gene = all_genes[i]
        last_char = gene[len(gene) - 1]
        if last_char == '+' or last_char == '-':
            all_genes[i] = gene[:len(gene) - 1]
    return all_genes


# remove genes in dict that are not in dataset
def filter_markers(adata, marker_dict):
    for category_name in marker_dict:
        markers = marker_dict[category_name]
        for key in markers:
            prior_size = len(markers[key])
            filtered_markers = set(markers[key]).intersection(adata.var.index)
            if len(filtered_markers) != prior_size:
                markers[key] = list(filtered_markers)
    return marker_dict


def get_markers(marker_paths):
    marker_dict = {}
    import json

    for marker_path in marker_paths:
        markers = {}
        with open(marker_path, 'rt') as f:
            marker_json = json.load(f)
            if 'title' in marker_json:
                marker_dict[marker_json['title']] = markers
                for cell_type in marker_json['cell_types']:
                    markers[cell_type['name']] = get_cell_type_genes(cell_type)
                    if 'subtypes' in cell_type:
                        for cell_sub_type in cell_type['subtypes']['cell_types']:
                            markers[cell_sub_type['name']] = get_cell_type_genes(cell_sub_type)
            else:
                key = os.path.splitext(os.path.basename(marker_path))[0]
                markers = marker_dict.get(key, {})
                marker_dict[key] = markers
                markers.update(marker_json)
    return marker_dict


def read_star_fusion_file(input_csv: str):
    """Load data from a STAR Fusion TSV file

    Parameters
    -----------

    input_csv : `str`
        The TSV file, gzipped or not, containing the STAR Fusion results.

    Returns
    -------
    """
    df = pd.read_csv(input_csv, sep='\t')
    df = df.sort_values('JunctionReadCount', ascending=False)
    df['value'] = True
    # df['value'] = df['value'].astype('category')
    df = df.drop_duplicates(['Cell', '#FusionName'], keep='first')
    # m = df.pivot_table('value', 'Cell', '#FusionName', fill_value=0)
    m = df.pivot('Cell', '#FusionName', 'value')
    m.fillna(False, inplace=True)
    return anndata.AnnData(obs=m)



import json
import os


def create_dataset_meta(path):
    result = {'id': path, 'url': path, 'name': os.path.splitext(os.path.basename(path))[0]}
    if os.path.basename(path).endswith('.json'):
        with open(path, 'rt') as f:
            result.update(json.load(f))
    return result


def write_json(json_data, json_path):
    with open(json_path, 'wt') as f:
        json.dump(json_data, f)


class LocalDbAPI:

    def __init__(self, paths):
        self.dataset_to_info = {}  # json_data, meta, json_path

        for path in paths:
            json_data = {}
            basename = os.path.splitext(path)[0]
            old_path = basename + '_filters.json'
            json_path = basename + '.json'
            if os.path.exists(old_path) and os.path.getsize(old_path) > 0:
                with open(old_path, 'rt') as f:
                    json_data['filters'] = json.load(f)

            if os.path.exists(json_path) and os.path.getsize(json_path) > 0:
                with open(json_path, 'rt') as f:
                    json_data.update(json.load(f))
            meta = create_dataset_meta(path)
            if 'filters' not in json_data:
                json_data['filters'] = {}
            if 'categories' not in json_data:
                json_data['categories'] = {}
            self.dataset_to_info[path] = dict(json_data=json_data, meta=meta, json_path=json_path)

    def server(self):
        return dict(mode='client')

    def user(self, email):
        return dict()

    def datasets(self, email):
        results = []
        for key in self.dataset_to_info:
            results.append(self.dataset_to_info[key]['meta'])
        return results

    def get_dataset(self, email, dataset_id, ensure_owner=False):
        result = self.dataset_to_info[dataset_id]['meta']
        result['id'] = dataset_id
        return result

    def category_names(self, email, dataset_id):
        results = []
        json_data = self.dataset_to_info[dataset_id]['json_data']
        categories = json_data['categories']
        for category_name in categories:
            category = categories[category_name]
            for category_key in category:
                r = dict(category=category_name, original=category_key, new=category[category_key]['new'])
                results.append(r)
        return results

    def upsert_category_name(self, email, category, dataset_id, original_name, new_name):
        json_data = self.dataset_to_info[dataset_id]['json_data']
        category_entity = json_data['categories'].get(category)
        if category_entity is None:
            category_entity = {}
            json_data['categories'][category] = category_entity
        if new_name == '':
            if original_name in category_entity:
                del category_entity[original_name]
        else:
            entity = dict(new=new_name)
            category_entity[original_name] = entity
            if dataset_id is not None:
                entity['dataset_id'] = dataset_id
            if email is not None:
                entity['email'] = email
        write_json(json_data, self.dataset_to_info[dataset_id]['json_path'])

    def dataset_filters(self, email, dataset_id):
        json_data = self.dataset_to_info[dataset_id]['json_data']
        results = []
        filters = json_data['filters']
        for key in filters:
            r = filters[key]
            r['id'] = key
            results.append(r)
        return results

    def delete_dataset_filter(self, email, dataset_id, filter_id):
        json_data = self.dataset_to_info[dataset_id]['json_data']
        del json_data['filters'][filter_id]
        write_json(json_data, self.dataset_to_info[dataset_id]['json_path'])

    def get_dataset_filter(self, email, dataset_id, filter_id):
        json_data = self.dataset_to_info[dataset_id]['json_data']
        return json_data['filters'][filter_id]

    def upsert_dataset_filter(self, email, dataset_id, filter_id, filter_name, filter_notes, dataset_filter):
        if filter_id is None:
            import uuid
            filter_id = str(uuid.uuid4())
        json_data = self.dataset_to_info[dataset_id]['json_data']
        entity = json_data['filters'].get(filter_id)
        if entity is None:
            entity = {}
            json_data['filters'][filter_id] = entity
        if filter_name is not None:
            entity['name'] = filter_name
        if dataset_filter is not None:
            entity['value'] = json.dumps(dataset_filter)
        if email is not None:
            entity['email'] = email
        if dataset_id is not None:
            entity['dataset_id'] = dataset_id
        if filter_notes is not None:
            entity['notes'] = filter_notes
        write_json(json_data, self.dataset_to_info[dataset_id]['json_path'])
        return filter_id



import subprocess
from codecs import open
from os import path

from setuptools import setup, find_packages
from setuptools.command.develop import develop
from setuptools.command.install import install


class PostDevelopCommand(develop):
    """Post-installation for development mode."""

    def run(self):
        subprocess.check_call(['npm', 'i'])
        subprocess.check_call(['npm', 'run-script', 'build'])
        develop.run(self)


class PostInstallCommand(install):
    """Post-installation for installation mode."""

    def run(self):
        subprocess.check_call(['npm', 'i'])
        subprocess.check_call(['npm', 'run-script', 'build'])
        install.run(self)


with open(path.join(path.abspath(path.dirname(__file__)), "README.rst"), encoding="utf-8") as f:
    long_description = f.read()

requires = [
        "anndata",
        "flask",
        "flask-compress",
        "fsspec",
        "gunicorn",
        "loompy",
        "natsort",
        "numpy",
        "pandas>=1.0",
        "pyarrow",
        "pymongo",
        "scipy"
]

setup(
    name="cirrocumulus",
    use_scm_version=True,
    setup_requires=['setuptools_scm'],
    # cmdclass={
    #         'develop': PostDevelopCommand,
    #         'install': PostInstallCommand,
    # },
    # version="2.0.2",
    description="Single-cell visualization tool",
    long_description=long_description,
    long_description_content_type="text/x-rst",
    url="https://github.com/klarman-cell-observatory/cirrocumulus",
    author="Joshua Gould",
    author_email='jgould@broadinstitute.org',
    classifiers=[
            "Intended Audience :: Developers",
            "Intended Audience :: Science/Research",
            "Natural Language :: English",
            "Operating System :: MacOS :: MacOS X",
            "Operating System :: Microsoft :: Windows",
            "Operating System :: POSIX :: Linux",
            "Programming Language :: Python :: 3",
            "Programming Language :: Python :: 3.6",
            "Programming Language :: Python :: 3.7",
            "Topic :: Scientific/Engineering :: Bio-Informatics",
            "Topic :: Scientific/Engineering :: Visualization"
    ],
    extras_require=dict(
        serve=[''],
        test=['pytest']
    ),
    keywords="single cell/nucleus genomics visualization",
    packages=find_packages(),
    include_package_data=True,
    install_requires=requires,
    python_requires=">=3.6",
    package_data={
    },
    entry_points={"console_scripts": ["cirro=cirrocumulus.__main__:main"]},
)
